/*
Copyright (C) 2010 Kevin Allen

This file is part of kacq.

kacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kacq.  If not, see <http://www.gnu.org/licenses/>.


 all the functions that needs to interact with the comedi drivers are here 
*/
#include "main.h"

int comedi_interface_init()
{
  // returns -1 if nothing is working
  // returns 0 if all ok
  // returns 1 if minor error
  int i;
  comedi_inter.sampling_rate=DEFAULT_SAMPLING_RATE;
  comedi_inter.buffer_size=MAX_BUFFER_LENGTH*MAX_DEVICES;
  comedi_inter.number_devices=0;
  comedi_inter.is_acquiring=0;
  comedi_inter.command_running=0;
  comedi_inter.data_offset_to_dat=-32768;
  comedi_inter.number_samples_read=0;
  comedi_inter.first_sample_in_buffer=0;
  comedi_inter.inter_acquisition_sleep_ms=5;
  comedi_inter.inter_acquisition_sleep_timespec=set_timespec_from_ms(comedi_inter.inter_acquisition_sleep_ms);
  comedi_inter. pause_restart_acquisition_thread_ms=100;
  comedi_inter.timespec_pause_restat_acquisition_thread=set_timespec_from_ms(comedi_inter.pause_restart_acquisition_thread_ms);

  if(comedi_dev_init(&comedi_inter.dev[0],"/dev/comedi0")!=0)
    {
      fprintf(stderr,"do not have a valid device to work with\n");
      return -1;
    }
  // device 0 is working so add one to num_device
  comedi_inter.number_devices++;


  if(MAX_DEVICES>1)
    {
      // try a second device.. this could be a loop to accomodate more than 2 devices.
      if ((comedi_dev_init(&comedi_inter.dev[1],"/dev/comedi1"))==0)
	{
	  printf("comedi_inter.dev[0].drive %s\n",comedi_inter.dev[0].driver);      
	  
	  // check if the two devices have the same drivers  
	  if((strcmp(comedi_inter.dev[0].driver,comedi_inter.dev[1].driver) !=0))
	    {
	      fprintf(stderr,"The two devices on the computer have different name or use different drivers\n");
	      fprintf(stderr,"device names: %s and %s\n driver names %s and %s",comedi_inter.dev[0].name,comedi_inter.dev[1].name,comedi_inter.dev[0].driver,comedi_inter.dev[1].driver);
	    }
	  else
	    { 
	      // we got all the information from device 2, all seems to work so add 1 to device counter
	      comedi_inter.number_devices++;
	    }
	}
    }
  // get the total number of channels available
  comedi_inter.number_channels=0;
  for (i=0;i<comedi_inter.number_devices;i++)
    {
      comedi_inter.number_channels+=comedi_inter.dev[i].number_channels_analog_input;
    }
  
  fprintf(stderr,"comedi_inter.number_channels: %d\n",comedi_inter.number_channels);
  comedi_interface_set_channel_list();
  comedi_inter.max_number_samples_in_buffer=MAX_BUFFER_LENGTH*MAX_DEVICES/comedi_inter.number_channels;
  // build the command to do acquisition
  comedi_interface_build_command();
  return 0;
}


int comedi_dev_init(struct comedi_dev *dev, char* file_name)
{
  // return 0 if all ok, -1 if problem
  
  int i;
  
  dev->file_name=file_name; // file to interface device 
  // try to open file
  if((dev->comedi_dev=comedi_open(dev->file_name))==NULL)
    {
      // that is serious problem because we need at least one working device
      fprintf(stderr, "error opening %s \n",dev->file_name);
      fprintf(stderr, "check that %s exists and that you have rw permission\n",dev->file_name);
      return -1;
    }
  // this device is valid so get the information about it structure
  dev->is_acquiring=0;
  dev->buffer_size=MAX_BUFFER_LENGTH;

  if((dev->name=comedi_get_board_name(dev->comedi_dev))==NULL)
    {
      fprintf(stderr, "problem getting the board name for %s\n", dev->file_name);
      return -1;
    }

  if((dev->driver=comedi_get_driver_name(dev->comedi_dev))==NULL)
    {
      fprintf(stderr, "problem getting the driver name for %s\n", dev->file_name);
      return -1;
    }

  if((dev->number_of_subdevices=comedi_get_n_subdevices(dev->comedi_dev))==-1)
    {
      fprintf(stderr, "problem getting the number of subdevices for %s\n", dev->file_name);
      return -1;
    }
  if((dev->subdevice_analog_input=comedi_find_subdevice_by_type(dev->comedi_dev,COMEDI_SUBD_AI,0))==-1)
    {
      fprintf(stderr, "problem finding a COMEDI_SUBD_AI subdevice on %s\n", dev->file_name);
      return -1;
    }

  if((dev->number_channels_analog_input=comedi_get_n_channels(dev->comedi_dev,dev->subdevice_analog_input))==-1)
    {
      fprintf(stderr, "problem getting the number of analog input channels on %s\n", dev->file_name);
      return -1;
    }
  
  if((dev->maxdata_input=comedi_get_maxdata(dev->comedi_dev,dev->subdevice_analog_input,0))==-1)
    {
      fprintf(stderr, "problem getting the maxdata from subdevice on %s\n", dev->file_name);
      return -1;
    }
  // set the channels from 0 to number_channels-1
  for (i=0; i < dev->number_channels_analog_input; i++)
    {
      dev->channel_list[i]=i;
    }
  dev->data_point_out_of_samples=0;
  return 0;
}

int comedi_interface_set_channel_list()
{
  // set the channel number on the user side
  int cumul=0;
  int i,j;
  fprintf(stderr,"comedi_interface_set_channel_list(), setting channel number for user side\n");
  for (i=0; i < comedi_inter.number_devices;i++)
    {
      for (j=0; j < comedi_inter.dev[i].number_channels_analog_input; j++)
	{
	  comedi_inter.channel_list[cumul+j]=cumul+j;
	}
      cumul=comedi_inter.dev[i].number_channels_analog_input;
    }
  return 0;
}

int comedi_interface_build_command()
{
  int i,j;
  // if only one device, configure to get all channels and start when command is run
  if (comedi_inter.number_devices==1)
    {
      fprintf(stderr, "building command for single device\n" );
      i=0;
      comedi_inter.dev[i].command.subdev         = comedi_inter.dev[i].subdevice_analog_input;
      comedi_inter.dev[i].command.flags          = 0;
      comedi_inter.dev[i].command.start_src      = TRIG_NOW;//  event to make the acquisition start
      comedi_inter.dev[i].command.start_arg      = 0;
      comedi_inter.dev[i].command.scan_begin_src = TRIG_TIMER;
      comedi_inter.dev[i].command.scan_begin_arg = 1e9/comedi_inter.sampling_rate;
      comedi_inter.dev[i].command.convert_src    = TRIG_TIMER;
      comedi_inter.dev[i].command.convert_arg    = 0;
      comedi_inter.dev[i].command.scan_end_src   = TRIG_COUNT;
      comedi_inter.dev[i].command.scan_end_arg   = comedi_inter.dev[i].number_channels_analog_input;
      comedi_inter.dev[i].command.stop_src       = TRIG_NONE;
      comedi_inter.dev[i].command.stop_arg       = 0;
      comedi_inter.dev[i].command.chanlist       = comedi_inter.dev[i].channel_list;
      comedi_inter.dev[i].command.chanlist_len   = comedi_inter.dev[i].number_channels_analog_input;
      for (j = 0 ; j < comedi_inter.dev[i].number_channels_analog_input; j++)
	{ // read x times the same channels so that the read functions returns quickly
	  
	  comedi_inter.dev[i].channel_list[j]=CR_PACK(comedi_inter.dev[i].channel_list[j],
						      comedi_inter.dev[i].range,
						      comedi_inter.dev[i].aref);
	}
      // check if the command can be done with the current device
      if (comedi_command_test(comedi_inter.dev[i].comedi_dev, &comedi_inter.dev[i].command)<0)
	{
	  fprintf(stderr, "error during comedi_command_test on %s\n", comedi_inter.dev[i].file_name);
	  return 1;
	}
    }
  if(comedi_inter.number_devices>1)
    {
      fprintf(stderr, "building command for master and slaves\n" );
      // set the card to work as master and slaves
      for (i=0; i <comedi_inter.number_devices;i++)
	{
	  if (i==0)
	    {      // this is the master card
	      fprintf(stderr, "building for master\n" );
	      comedi_t_enable_master(comedi_inter.dev[i].comedi_dev);
	      comedi_inter.dev[i].command.subdev         = comedi_inter.dev[i].subdevice_analog_input;
	      comedi_inter.dev[i].command.flags          = 0;
	      comedi_inter.dev[i].command.start_src      = TRIG_NOW;//  event to make the acquisition start
	      comedi_inter.dev[i].command.start_arg      = 0;
	      comedi_inter.dev[i].command.scan_begin_src = TRIG_TIMER;
	      comedi_inter.dev[i].command.scan_begin_arg = 1e9/comedi_inter.sampling_rate;
	      comedi_inter.dev[i].command.convert_src    = TRIG_TIMER;
	      comedi_inter.dev[i].command.convert_arg    = 0;
	      comedi_inter.dev[i].command.scan_end_src   = TRIG_COUNT;
	      comedi_inter.dev[i].command.scan_end_arg   = comedi_inter.dev[i].number_channels_analog_input;
	      comedi_inter.dev[i].command.stop_src       = TRIG_NONE;
	      comedi_inter.dev[i].command.stop_arg       = 0;
	      comedi_inter.dev[i].command.chanlist       = comedi_inter.dev[i].channel_list;
	      comedi_inter.dev[i].command.chanlist_len   = comedi_inter.dev[i].number_channels_analog_input;
	      for (j = 0 ; j < comedi_inter.dev[i].number_channels_analog_input; j++)
		{ // read x times the same channels so that the read functions returns quickly
		  
		  comedi_inter.dev[i].channel_list[j]=CR_PACK(comedi_inter.dev[i].channel_list[j],
							      comedi_inter.dev[i].range,
							      comedi_inter.dev[i].aref);
		}
	      // check if the command can be done with the current device
	      if (comedi_command_test(comedi_inter.dev[i].comedi_dev, &comedi_inter.dev[i].command)<0)
		{
		  fprintf(stderr, "error during comedi_command_test on %s\n", comedi_inter.dev[i].file_name);
		  return 1;
		}
	    }
	  else
	    {
	      // set the slave cards
	      fprintf(stderr, "building for slave\n" );
	      comedi_t_enable_slave(comedi_inter.dev[i].comedi_dev);
	      comedi_inter.dev[i].command.subdev         = comedi_inter.dev[i].subdevice_analog_input;
	      comedi_inter.dev[i].command.flags          = 0;
	      comedi_inter.dev[i].command.start_src      = TRIG_EXT;//  start from master
	      comedi_inter.dev[i].command.start_arg      = CR_EDGE | NI_EXT_RTSI(0);
	      comedi_inter.dev[i].command.scan_begin_src = TRIG_TIMER;
	      comedi_inter.dev[i].command.scan_begin_arg = 1e9/comedi_inter.sampling_rate;
	      comedi_inter.dev[i].command.convert_src    = TRIG_EXT;
	      comedi_inter.dev[i].command.convert_arg    = CR_INVERT | CR_EDGE | NI_EXT_RTSI(2);
	      comedi_inter.dev[i].command.scan_end_src   = TRIG_COUNT;
	      comedi_inter.dev[i].command.scan_end_arg   = comedi_inter.dev[i].number_channels_analog_input;
	      comedi_inter.dev[i].command.stop_src       = TRIG_NONE;
	      comedi_inter.dev[i].command.stop_arg       = 0;
	      comedi_inter.dev[i].command.chanlist       = comedi_inter.dev[i].channel_list;
	      comedi_inter.dev[i].command.chanlist_len   = comedi_inter.dev[i].number_channels_analog_input;
	      for (j = 0 ; j < comedi_inter.dev[i].number_channels_analog_input; j++)
		{ // read x times the same channels so that the read functions returns quickly
		  
		  comedi_inter.dev[i].channel_list[j]=CR_PACK(comedi_inter.dev[i].channel_list[j],
							      comedi_inter.dev[i].range,
							      comedi_inter.dev[i].aref);
		}
	      // check if the command can be done with the current device
	      if (comedi_command_test(comedi_inter.dev[i].comedi_dev, &comedi_inter.dev[i].command)<0)
		{
		  fprintf(stderr, "error during comedi_command_test on %s\n", comedi_inter.dev[i].file_name);
		  return 1;
		}
	    }
	}
    }
  return 0;
}

int comedi_t_enable_master(comedi_t *dev)
{

  fprintf(stderr, "enabling master\n" );
  comedi_insn   configCmd;
  lsampl_t      configData[2];
  int           ret;
  static const unsigned rtsi_subdev = 10;
  static const unsigned rtsi_clock_line = 7;

  /* Route RTSI clock to line 7 (not needed on pre-m-series boards since their
     clock is always on line 7). */
  memset(&configCmd, 0, sizeof(configCmd));
  memset(&configData, 0, sizeof(configData));
  configCmd.insn = INSN_CONFIG;
  configCmd.subdev = rtsi_subdev;
  configCmd.chanspec = rtsi_clock_line;
  configCmd.n = 2;
  configCmd.data = configData;
  configCmd.data[0] = INSN_CONFIG_SET_ROUTING;
  configCmd.data[1] = NI_RTSI_OUTPUT_RTSI_OSC;
  ret = comedi_do_insn(dev, &configCmd);
  if(ret < 0){
    comedi_perror("comedi_do_insn: INSN_CONFIG");
    return(1);
  }
  // Set clock RTSI line as output
  ret = comedi_dio_config(dev, rtsi_subdev, rtsi_clock_line, INSN_CONFIG_DIO_OUTPUT);
  if(ret < 0){
    comedi_perror("comedi_dio_config");
    return(1);
  }
  
  /* Set routing of the 3 main AI RTSI signals and their direction to output.
     We're reusing the already initialized configCmd instruction here since
     it's mostly the same. */
  configCmd.chanspec = 0;
  configCmd.data[1] =  NI_RTSI_OUTPUT_ADR_START1;
  ret = comedi_do_insn(dev, &configCmd);
  if(ret < 0){
    comedi_perror("comedi_do_insn: INSN_CONFIG");
    return(1);
  }
  ret = comedi_dio_config(dev, rtsi_subdev, 0, INSN_CONFIG_DIO_OUTPUT);
  if(ret < 0){
    comedi_perror("comedi_dio_config");
    return(1);
  }
  
  configCmd.chanspec = 1;
  configCmd.data[1] =  NI_RTSI_OUTPUT_ADR_START2;
  ret = comedi_do_insn(dev, &configCmd);
  if(ret < 0){
    comedi_perror("comedi_do_insn: INSN_CONFIG");
    return(1);
  }
  ret = comedi_dio_config(dev, rtsi_subdev, 1, INSN_CONFIG_DIO_OUTPUT);
  if(ret < 0){
    comedi_perror("comedi_dio_config");
    return(1);
  }
  
  configCmd.chanspec = 2;
  configCmd.data[1] =  NI_RTSI_OUTPUT_SCLKG;
  ret = comedi_do_insn(dev, &configCmd);
  if(ret < 0){
    comedi_perror("comedi_do_insn: INSN_CONFIG");
    return(1);
  }
  ret = comedi_dio_config(dev, rtsi_subdev, 2, INSN_CONFIG_DIO_OUTPUT);
  if(ret < 0){
    comedi_perror("comedi_dio_config");
    return(1);
  }
  return 0;
}
int comedi_t_enable_slave(comedi_t *dev)
{
  fprintf(stderr, "enabling slave\n" );
  comedi_insn   configCmd;
  lsampl_t      configData[3];
  int           ret;
  static const unsigned rtsi_subdev = 10;
  static const unsigned rtsi_clock_line = 7;
  
  memset(&configCmd, 0, sizeof(configCmd));
  memset(&configData, 0, sizeof(configData));
  configCmd.insn = INSN_CONFIG;
  configCmd.subdev = rtsi_subdev;
  configCmd.chanspec = 0;
  configCmd.n = 3;
  configCmd.data = configData;
  configCmd.data[0] = INSN_CONFIG_SET_CLOCK_SRC;
  configCmd.data[1] = NI_MIO_PLL_RTSI_CLOCK(rtsi_clock_line);
  configCmd.data[2] = 100;	/* need to give it correct external clock period */
  ret = comedi_do_insn(dev, &configCmd);
  if(ret < 0){
    comedi_perror("comedi_do_insn: INSN_CONFIG");
    return(1);
  }
  /* configure RTSI clock line as input */
  ret = comedi_dio_config(dev, rtsi_subdev, rtsi_clock_line, INSN_CONFIG_DIO_INPUT);
  if(ret < 0){
    comedi_perror("comedi_dio_config");
    return(1);
  }
  /* Configure RTSI lines we are using for AI signals as inputs. */
  ret = comedi_dio_config(dev, rtsi_subdev, 0, INSN_CONFIG_DIO_INPUT);
  if(ret < 0){
    comedi_perror("comedi_dio_config");
    return(1);
  }
  ret = comedi_dio_config(dev, rtsi_subdev, 1, INSN_CONFIG_DIO_INPUT);
  if(ret < 0){
    comedi_perror("comedi_dio_config");
    return(1);
  }
  ret = comedi_dio_config(dev, rtsi_subdev, 2, INSN_CONFIG_DIO_INPUT);
  if(ret < 0){
    comedi_perror("comedi_dio_config");
    return(1);
  }
  return 0;
}

int comedi_interface_run_command()
{
  // because of the master/slave arrangment and because card 0 is the master
  // we need to start device 0 last
  fprintf(stderr, "comedi_interface_run_command()\n");
  int i;
  for (i=comedi_inter.number_devices-1;i>=0;i--)
    {
      if ((comedi_command(comedi_inter.dev[i].comedi_dev, &comedi_inter.dev[i].command)) < 0) 
	{
	  fprintf(stderr, "comedi_command failed to start for %s\n",comedi_inter.dev[i].file_name);
	  return (1);
	}
    }
  comedi_inter.command_running=1;
  return 0;
}
int comedi_interface_stop_command()
{
  
  fprintf(stderr, "comedi_interface_stop_command()\n");
  int i;
  for (i=0; i <comedi_inter.number_devices;i++)
    {
      if ((comedi_cancel(comedi_inter.dev[i].comedi_dev, comedi_inter.dev[i].subdevice_analog_input)) < 0) 
	{
	  fprintf(stderr, "comedi_cancel failed to stop for %s\n",comedi_inter.dev[i].file_name);
	  return (1);
	}
    }
  comedi_inter.command_running=0;
  return 0;
}



int comedi_interface_get_data_from_devices()
{
  int i,j,k;
  int samples_move_back;
  int channel_device_offset;
  // call function to get the data from device
  for (i=0; i <comedi_inter.number_devices;i++)
    {
      if (comedi_dev_read_data(&comedi_inter.dev[i])==-1)
	{
	  fprintf(stderr, "error reading from comedi device %s\n",comedi_inter.dev[i].file_name);
	  comedi_inter.command_running=0;
	  return 1;
	}
    }
  
  // get the number of complete samples that can be put in interface buffer
  // which is the smallest number of samples loaded by one of the devices
  for (i=0; i <comedi_inter.number_devices;i++)
    {
      if (i==0)
	{
	  comedi_inter.min_samples_read_from_devices=comedi_inter.dev[i].samples_read;
	}  
      else
	{
	  if(comedi_inter.min_samples_read_from_devices>comedi_inter.dev[i].samples_read)
	    {
	      comedi_inter.min_samples_read_from_devices=comedi_inter.dev[i].samples_read;
	    }
	}
    }
  //  fprintf(stderr, "comedi_interface_get_data_from_device: min_samples in devices buffer: %d\n",comedi_inter.min_samples_read_from_devices);



  // adjust the number of unused data for each device, they will not be overwritten on next call to comedi_dev_read_data
  // and will be available on the next call to this function
  for (i=0; i <comedi_inter.number_devices;i++)
    {
      comedi_dev_adjust_data_point_out_of_samples(&comedi_inter.dev[i], comedi_inter.min_samples_read_from_devices);
    }
 
 
  // use to prevent access to comedi_inter.buffer_data while modifying it
  pthread_mutex_lock( &mutex_comedi_interface_buffer );

  
  //fprintf(stderr, "comedi_interface_get_data_from_device: starting transfer\n");

  // check if we need to make space at the end of the buffer for the new data
  if(comedi_inter.number_samples_read+comedi_inter.min_samples_read_from_devices>comedi_inter.max_number_samples_in_buffer)
    {
      //fprintf(stderr, "comedi_interface_get_data_from_device: moving back data in interface buffer\n");
      //fprintf(stderr, "comedi_interface_get_data_from_device: number_samples_read:%ld max_number_samples_in_buffer:%d\n",comedi_inter.number_samples_read,comedi_inter.max_number_samples_in_buffer);

      // we need to make some space in the buffer, some data will be thrown away
      if (comedi_inter.number_samples_read<comedi_inter.max_number_samples_in_buffer)
	{
	  // buffer not full yet, we only need to move the old data a bit
	  samples_move_back=comedi_inter.min_samples_read_from_devices-(comedi_inter.max_number_samples_in_buffer-comedi_inter.number_samples_read);
	}
      else
	{
	  // buffer already full, move back by the size of the new data
	  samples_move_back=comedi_inter.min_samples_read_from_devices;
	}

      // move back loop to free the end of the buffer
      //fprintf(stderr, "comedi_interface_get_data_from_device: samples_move_back: %d\n",samples_move_back);
      comedi_inter.data_points_move_back=(comedi_inter.max_number_samples_in_buffer-samples_move_back)*comedi_inter.number_channels;
      comedi_inter.offset_move_back=samples_move_back*comedi_inter.number_channels;
      for(i=0;i<comedi_inter.data_points_move_back;i++)
	{
	  comedi_inter.buffer_data[i]=comedi_inter.buffer_data[i+comedi_inter.offset_move_back];
	}
      
      // put the data directly at the end of the buffer
      channel_device_offset=0;
      for (i=0; i < comedi_inter.number_devices;i++)
	{
	  if(i>0)
	    {
	      channel_device_offset+=comedi_inter.dev[i].number_channels_analog_input;
	    }
	  //fprintf(stderr, "comedi_interface_get_data_from_device: adding data at end from sample: %d \n",(comedi_inter.max_number_samples_in_buffer-comedi_inter.min_samples_read_from_devices)*comedi_inter.number_channels);
	  for(j=0;j<comedi_inter.min_samples_read_from_devices;j++)
	    {
	      for(k=0;k<comedi_inter.dev[i].number_channels_analog_input;k++)
		{
		  comedi_inter.buffer_data[(comedi_inter.max_number_samples_in_buffer-comedi_inter.min_samples_read_from_devices+j)*comedi_inter.number_channels+k+channel_device_offset]=
		    comedi_inter.dev[i].buffer_data[j*comedi_inter.dev[i].number_channels_analog_input+k]+comedi_inter.data_offset_to_dat;
		}
	    }
	}
      // update the number of the first sample in buffer
      comedi_inter.first_sample_in_buffer+=samples_move_back;
    }
  else
    {
      //fprintf(stderr, "comedi_interface_get_data_from_device: adding data after the last data\n");
      // there is enough space after the old data
      // just add the new data after the old data in the buffer
      channel_device_offset=0;
      for (i=0; i < comedi_inter.number_devices;i++)
	{
	  if(i>0)
	    {
	      channel_device_offset+=comedi_inter.dev[i].number_channels_analog_input;
	    }    
	  //fprintf(stderr, "comedi_interface_get_data_from_device: adding data from sample: %ld \n",comedi_inter.number_samples_read);
	  for(j=0;j<comedi_inter.min_samples_read_from_devices;j++) // for all samples
	    {
	      for(k=0;k<comedi_inter.dev[i].number_channels_analog_input;k++)
		{                           // [number samples already in buffer+added samples * total number of channels       +   chan_no   +   offset of the card]
		  comedi_inter.buffer_data[ (comedi_inter.number_samples_read+j)               * comedi_inter.number_channels   +   k         +   channel_device_offset]=

		    //                              [added_samples *   number_channels on that device + chan_no]       comedi_inter.data_offset_to_dat
		    comedi_inter.dev[i].buffer_data[     j        *   comedi_inter.dev[i].number_channels_analog_input+k]+comedi_inter.data_offset_to_dat;
		}
	    }
	}
    }
  comedi_inter.number_samples_read+=comedi_inter.min_samples_read_from_devices;
  // allow other thread to read comedi_inter.buffer_data and access other variables
  pthread_mutex_unlock(&mutex_comedi_interface_buffer );
  return 0;
}  
int comedi_dev_adjust_data_point_out_of_samples(struct comedi_dev *dev,int number_samples_transfered_to_inter_buffer)
{
  //  fprintf(stderr,"comedi_dev_adjust_data_point_out_of_samples: number_samples_transfered_to_inter_buffer: %d\n", number_samples_transfered_to_inter_buffer);
  int unused_samples;
  unused_samples=dev->samples_read-number_samples_transfered_to_inter_buffer; 
  dev->data_point_out_of_samples+=unused_samples*dev->number_channels_analog_input;
  dev->samples_read=number_samples_transfered_to_inter_buffer;
  dev->cumulative_samples_read-=unused_samples;
  //fprintf(stderr,"comedi_dev_adjust_data_point_out_of_samples: unused_samples:%d samples_read: %d\n",unused_samples,dev->samples_read);
  return 0;
}
int comedi_dev_read_data(struct comedi_dev *dev)
{
  int i;
  // move back incomplete sample data and complete samples not used to the beginning of the buffer
  for (i=0;i<dev->data_point_out_of_samples;i++)
    {
      dev->buffer_data[i]=dev->buffer_data[dev->samples_read * dev->number_channels_analog_input +i];
    }
  
  // move the pointer to avoid overwritting the incomplete sample data
  dev->pointer_buffer_data=dev->buffer_data+dev->data_point_out_of_samples;

  // read new data from device
  dev->read_bytes=read(comedi_fileno(dev->comedi_dev),dev->pointer_buffer_data,sizeof(dev->buffer_data)-(sizeof(sampl_t)*dev->data_point_out_of_samples));
  
  // check if read was successful
  if (dev->read_bytes<0)// reading failed
    {
      fprintf(stderr, "error reading from comedi device %s\n",dev->file_name);
      return 1;
    }
  else // read was ok
    {
      // get number of samples read
      dev->samples_read=(dev->read_bytes + (dev->data_point_out_of_samples*sizeof(sampl_t))) / (sizeof(sampl_t)*dev->number_channels_analog_input);
      dev->cumulative_samples_read+=dev->samples_read;
      dev->data_point_out_of_samples=(dev->read_bytes/sizeof(sampl_t) + dev->data_point_out_of_samples) % dev->number_channels_analog_input;
    }
  //  fprintf(stderr, "comedi_dev_read_data(): samples read: %d, data_point_out_of_samples: %d\n",dev->samples_read,dev->data_point_out_of_samples);
  return 0;
}

int comedi_interface_clear_current_acquisition_variables()
{
  int i=0;
  fprintf(stderr, "comedi_interface_clear_current_acquisition_variables()\n");
  comedi_inter.number_samples_read=0;
  comedi_inter.first_sample_in_buffer=0;

  for(i=0;i<comedi_inter.number_devices;i++)
    {
      comedi_dev_clear_current_acquisition_variables(&comedi_inter.dev[i]);
    }
  return 0;
}
int comedi_dev_clear_current_acquisition_variables(struct comedi_dev *dev)
{
  dev->read_bytes=0;
  dev->samples_read=0;
  dev->data_point_out_of_samples=0;
  dev->cumulative_samples_read=0;
  return 0;
}
