/****************************************************************
Copyright (C) 2010 Kevin Allen

This file is part of kacq.

kacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kacq.  If not, see <http://www.gnu.org/licenses/>.

declaration of the main structures and functions used in kacq

gcc -Wall -g -o kacq main.c callback.c comedi_code.c recording.c  -export-dynamic `pkg-config --cflags --libs gtk+-2.0 gtkdatabox`  -lcomedi -lm -lrt -pthread
****************************************************************/
#include <comedi.h> // for the driver
#include <comedilib.h> // for the driver API
#include <stdio.h>
#include <fcntl.h> // for file operations
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h> // for the nanosleep
#include <gtk/gtk.h> // for the widgets
#include <pthread.h> // to be able to create threads
#include <glib/gprintf.h>
#include <gtkdatabox.h>
#include <gtkdatabox_points.h>
#include <gtkdatabox_ruler.h>
#include <math.h>
#include <pwd.h> // to get the home directory as default directory
#include <stdlib.h>


#define _FILE_OFFSET_BITS 64 // to have files larger than 2GB
#define MAX_CHANNELS 64
#define MAX_DEVICES 2
#define MAX_CHAR 255
#define NSEC_PER_SEC (1000000000) // The number of nsecs per sec
#define MAX_BUFFER_LENGTH 40000 // buffer length for each comedi_dev
#define DEFAULT_SAMPLING_RATE 20000 
#define FACTOR_RECORDING_BUFFER 4 // factor by which the recording_interface buffer > comedi_interface buffer */

#define OSCILLOSCOPE_GROUPS 10 // number of groups of channels that can be displayed
#define MAX_CHANNELS_PER_GROUP 8
#define MAX_TIME_SEC_IN_OSCILLOSCOPE_PAGE 1
#define FACTOR_OSCILLATORY_BUFFER 10

/***************************************************
The following variables are for the program threads
Acquisition thread is stopped by setting comedi_interface.is_acquiring to 0
Recording thread is stopped by setting the recording_interface.is_recording to 0
****************************************************/
pthread_t recording_thread;
int recording_thread_id;
char* t;
pthread_t acquisition_thread;
int acquisition_thread_id;
/* mutex to prevent thread to read from a buffer that is being modified
 all parts of code reading or writing to comedi_inter.buffer_data should
 be surrounded by 
 pthread_mutex_lock( &mutex1 ); and   pthread_mutex_unlock( &mutex1 );
*/
pthread_mutex_t mutex_comedi_interface_buffer;





/***********************************************************************************
 structure that holds all the gui widgets that we need to control during execution
***********************************************************************************/
struct all_widget
{ // see kacq.glade file to know where these appear
  GtkWidget *window;  // main window
  GtkWidget *vbox1; // main vbox in the main window, to put the gtkdatabox

  GtkWidget *test_label;
  GtkDataboxGraph *graph;
  GtkWidget *table;
  GtkWidget *box;
  GdkColor color;
 

  GtkWidget *about_dlg; // about dialog
  GtkWidget *preferences_dlg;
  GtkWidget *toolbar;
  GtkWidget *dev1_name_label; // to show user the name of device
  GtkWidget *dev2_name_label; // to show user the driver
  GtkWidget *dev1_driver_label;
  GtkWidget *dev2_driver_label;
  GtkWidget *num_devices_detected_label;
  GtkWidget *sampling_rate_entry;
  GtkWidget *num_available_channels_label;
  GtkWidget *num_channels_device_1_label;
  GtkWidget *num_channels_device_2_label;
  GtkWidget *preferences_channel_vbox; // to display the channel information
  GtkWidget *file_name_entry; // filebase of the file name
  GtkWidget *trial_spinbutton; // index following filebase for file name
  GtkWidget *statusbar; // index following filebase for file name   
  GtkWidget *channel_view; // treeview to select the channels to record
  GtkListStore  *channel_store; // model to fill up the treview to select the channels to record
};
struct all_widget widgets;

// enumeration to identify the columns in the model to select the channels to record
enum
{
  COL_NO = 0,
  COL_DEVICE,
  COL_SELECT,
  NUM_COLS
} ;



/******************************************************
 structure to hold the information regarding an AD card
Notes:
Acquisition always reads from all available channels

The selection of which channels are displayed or
recorded will not affect comedi_dev code

Each device has its own buffer to read data, but 
the data should always be accessed via the 
comedi_interface buffer
*******************************************************/
struct comedi_dev
{
  comedi_t *comedi_dev; // device itself
  char *file_name; // file name for access to device
  const char *name; // name of the card
  const char *driver; // name of the comedi driver
  int number_of_subdevices;
  int subdevice_analog_input; // id of analog input subdev
  int number_channels_analog_input;
  int maxdata_input;
  comedi_range * range_input;
  double voltage_max_input;
  int aref;
  int range;
  int buffer_size;
  sampl_t buffer_data[MAX_BUFFER_LENGTH];
  sampl_t* pointer_buffer_data; // to accomodate for incomplete read sample
  int read_bytes;
  int samples_read;
  int data_point_out_of_samples; // because read operation returns incomplete samples
  long int cumulative_samples_read;
  comedi_cmd command; 
  unsigned int channel_list[MAX_CHANNELS]; // channel number for the comedi side
  int is_acquiring;
};


/*********************************************************
structure to hold the information about all cards together

comed_interface has a buffer with the data from all
devices in it. The recording and gui part of the program
should only interface with comedi_interface and never directly 
with the comedi_dev.



!!!!!!!! URGENT TO DO !!!!!!!!!!!!!!!!!!!1
the interface buffer has to be modified so that we never
move data in the buffer but only overwrite the old data
that will make the read and write to buffer more complex
but should be much more efficent. only two functions need 
to be aware of that, the one putting data in and the one
taking data out.

*********************************************************/
struct comedi_interface
{
  int is_acquiring;
  int command_running;
  int number_devices; // counter of valid devices
  int sampling_rate; // sampling rate of recording
  int number_channels; // aggregate of two cards
  int max_number_samples_in_buffer;
  int min_samples_read_from_devices; // if compare the different devices
  int data_offset_to_dat;
  struct comedi_dev dev[MAX_DEVICES]; // hold the info about each device
  unsigned int channel_list[MAX_CHANNELS*MAX_DEVICES]; // hold the channel number user side
  int buffer_size;
  short int buffer_data[MAX_BUFFER_LENGTH*MAX_DEVICES];// buffer to get data from comedi devices
  int data_points_move_back;
  int offset_move_back;
  unsigned long int number_samples_read; // total samples read, 
  unsigned long int first_sample_in_buffer; // sample no of the first sample in buffer
  struct timespec inter_acquisition_sleep_timespec; // between read operations to driver
  double inter_acquisition_sleep_ms;
  struct timespec timespec_pause_restat_acquisition_thread; // to allow acquisition to complete
  double pause_restart_acquisition_thread_ms;
  struct timespec req;
  };


/*********************************************************
Structure holding variables for the recording process
********************************************************/
struct recording_interface
{
  int is_recording;
  FILE *file;
  gchar* file_name; // directory + file name from gui
  gchar* directory; // 
  off_t file_size; // to check if the file size makes sense
  off_t predicted_file_size; // to check the size of file at the end
  int number_of_channels_to_save; // number of channels that will be saved
  unsigned int channel_list[MAX_CHANNELS*MAX_DEVICES]; // list of channels that will be saved
  int rec_buffer_size;
  int max_samples_in_buffer;
  short int rec_buffer_data[MAX_BUFFER_LENGTH*MAX_DEVICES*FACTOR_RECORDING_BUFFER];
  struct timespec inter_recording_sleep_timespec; // sleeping time between attemping to save
  struct timespec req;
  double inter_recording_sleep_ms;
  // following variables that changes during recording process
  int new_samples_in_buffer;
  unsigned long int number_samples_saved; // this is the 0-based index of first sample in the recording buffer
  // context for statusbar, the status bar will show details on the recording currently in process
  guint statusbar_context_id;
  guint statusbar_message_id;
};


/*********************************************************
Structure holding variables for the oscilloscope
********************************************************/
struct oscilloscope_interface
{
  // a page is the data required to fill one screen of data, it has all the channels in it
  int number_pages_in_buffer; // will vary according to the time resolution of oscilloscope
                              // with slow resolution, a page contains more data, so fewer pages in total

  int number_groups; // channels are always displayed by group of x channels 
  int max_channels_per_group; // to avoid having to much data to display on the oscilloscope
   
  int number_channels;
  // the osc_buffer has a copy of all the recorded channels, that will make simpler to code
  // it is devided into a number of pages, each page contains the data to fill a display
  short int osc_buffer[MAX_BUFFER_LENGTH*MAX_DEVICES*FACTOR_OSCILLATORY_BUFFER];
  int buffer_size;

  double seconds_per_page; // second per page
  int samples_per_page;
  int page_size;
};





struct comedi_interface comedi_inter;
struct recording_interface recording_inter;
struct oscilloscope_interface osc_inter;

GtkBuilder *builder; // to build the interface from glade



/***********************
 defined in callback.c
***********************/
void on_window_destroy (GtkObject *object, gpointer user_data);
// menuitems
void on_open_menuitem_activate(GtkObject *object, gpointer user_data);
void on_save_menuitem_activate(GtkObject *object, gpointer user_data);
void on_save_as_menuitem_activate(GtkObject *object, gpointer user_data);
void on_quit_menuitem_activate(GtkObject *object, gpointer user_data);
void on_about_menuitem_activate(GtkObject *object, gpointer user_data);
void on_preferences_menuitem_activate(GtkObject *object, gpointer user_data);
// about dialot
void on_about_dialog_delete_event(GtkObject *object, gpointer user_data);
// preferences dialog
void on_preferences_dialog_delete_event(GtkObject *object, gpointer user_data);
void on_ok_preferences_button_activate(GtkObject *object, gpointer user_data);
void cell_toggled_callback (GtkCellRendererToggle *cell,gchar *path_string, gpointer user_data);
// toolbar buttons
void on_previous_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_rewind_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_play_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_stop_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_forward_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_next_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_record_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_gain_increase_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_gain_decrease_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_time_increase_toolbutton_clicked(GtkObject *object, gpointer user_data);
void on_time_decrease_toolbutton_clicked(GtkObject *object, gpointer user_data);

// functions called at startup, to initialize variables
int init_window(); // to complete building the interface
int set_device_labels(); // to set the labels in the comedi preferences dialog
int build_channel_tree_view(); // build the tree view to dispaly the channels to record
                               // in the preferences dialog
void create_and_fill_model();

// functions used to get the time of some operations
struct timespec set_timespec_from_ms(double milisec);
struct timespec diff(struct timespec start, struct timespec end);
int microsecond_from_timespec(struct timespec duration);


/************************
defined in recording.c
************************/
int recording_interface_init(struct recording_interface* rec);
int recording_interface_open_file(struct recording_interface* rec);
int recording_interface_get_data(struct recording_interface* rec, struct comedi_interface* com);
int recording_interface_save_data(struct recording_interface* rec);
int recording_interface_close_file(struct recording_interface* rec);
int recording_interface_check_file_size(struct recording_interface* rec);
int recording_interface_clear_current_recording_variables(struct recording_interface* rec);
int recording_interface_get_channels_from_channel_store(struct recording_interface* rec,struct comedi_interface* com, GtkListStore  *channel_store);

/*************************
 define in comedi_code.c
**************************/
int comedi_interface_init(); // get the information on the AD cards installed
int comedi_interface_set_channel_list();
int comedi_interface_build_command();
int comedi_interface_run_command();
int comedi_interface_stop_command();
int comedi_interface_get_data_from_devices();
int comedi_interface_clear_current_acquisition_variables();
int comedi_dev_init(struct comedi_dev *dev, char* file_name);
int comedi_dev_read_data(struct comedi_dev *dev);
int comedi_dev_adjust_data_point_out_of_samples(struct comedi_dev *dev,int number_samples_transfered_to_inter_buffer);
int comedi_dev_clear_current_acquisition_variables(struct comedi_dev *dev);
int comedi_t_enable_master(comedi_t *dev);
int comedi_t_enable_slave(comedi_t *dev);



/********************************
defined in oscilloscope.c
********************************/
int oscilloscope_interface_init(struct oscilloscope_interface* osc);






/***************************************************
functions run at by the thread starter
the thread is killed when function exits
body of the function is in callback.c
***************************************************/
void *recording(void* thread_id);
void *acquisition(void* thread_id);
