/* 
Copyright (C) 2010 Kevin Allen

This file is part of kacq.

kacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kacq.  If not, see <http://www.gnu.org/licenses/>.

all the functions for the recording part of kacq 
*/
#include "main.h"

int recording_interface_init(struct recording_interface* rec)
{
  /* initialize the recording interface, by default it records all channels */
  int i;
  rec->is_recording=0;
  rec->number_of_channels_to_save=MAX_CHANNELS;
  
  // set the home directory as the default directory
  struct passwd *p;
  char *username=getenv("USER");
  p=getpwnam(username);
  rec->directory=strcat(p->pw_dir,"/");
  

  
  rec->rec_buffer_size=MAX_BUFFER_LENGTH*MAX_DEVICES*FACTOR_RECORDING_BUFFER;
  rec->max_samples_in_buffer=rec->rec_buffer_size/rec->number_of_channels_to_save;
  rec->new_samples_in_buffer=0;
  rec->file_size=0;
  rec->inter_recording_sleep_ms=20;
  rec->inter_recording_sleep_timespec=set_timespec_from_ms(rec->inter_recording_sleep_ms);
  rec->number_samples_saved=0;
  for (i=0;i<rec->number_of_channels_to_save;i++)
    {
      rec->channel_list[i]=i;
    }
  return 0;
}

int recording_interface_open_file(struct recording_interface* rec)
{
  fprintf(stderr, "recording_interface_open_file()\n");
  rec->file=fopen(rec->file_name,"w");
  if (rec->file==NULL)
    {
      fprintf(stderr,"error opening %s in recording_interface_open_file()\n",rec->file_name);
      return -1;
    }
  return 0;
}
int recording_interface_get_data(struct recording_interface* rec, struct comedi_interface*com)
{
  /* this function takes new available samples from the comedi_interface and put them into the recording_interface buffer 
     if the buffer is more than half full, then save the data to disk
   */
  int index_acq, index_rec,num_samples_to_transfer,i,j;
  // check if the comedi device has new samples that are not in current recording  buffer
  if(com->number_samples_read > rec->number_samples_saved+rec->new_samples_in_buffer)
    {
      // prevent other threads from changing comedi_interface_buffer
      pthread_mutex_lock( &mutex_comedi_interface_buffer );
      
      // find the index in the comedi_interface buffer of the first new samples we want to transfer to recording buffer
      index_acq=((rec->number_samples_saved+rec->new_samples_in_buffer)-com->first_sample_in_buffer)*  com->number_channels; // the no of first new sample
      
      // find the index in the recording buffer for the new samples coming from comedi_interface buffer
      index_rec=(rec->new_samples_in_buffer)*rec->number_of_channels_to_save;
      
      // calculate the number of samples to transfer
      if (com->number_samples_read -  rec->number_samples_saved < rec->max_samples_in_buffer)
	{
	  // there is space for all the samples available from comedi_interface
	  num_samples_to_transfer=com->number_samples_read - rec->number_samples_saved;
	}
      else
	{
	  // there is only space for part of samples available
	  num_samples_to_transfer=rec->max_samples_in_buffer-rec->new_samples_in_buffer;
	}
      
      if(index_acq<0)
	{
	  // that means that we have lost some data. Critical error. Recording should end now
	  fprintf(stderr,"recording_interface_get_data: the new data needed are no longer in comedi_interface buffer.\n Recording stop here\n");
	  rec->is_recording=0;
	  return -1;
	}
      if(index_acq+num_samples_to_transfer*com->number_channels>com->buffer_size)
	{
	  fprintf(stderr,"attempt to read comedi_interface->buffer_data past the buffer end.\n Buffer size: %d, reading from %d to %d\n",com->buffer_size,index_acq,index_acq+num_samples_to_transfer*com->number_channels);
	  rec->is_recording=0;
	  return -1;
	}
           
      // transfer the data
      for(i=0;i<num_samples_to_transfer;i++)
	{
	  for(j=0;j<rec->number_of_channels_to_save;j++)
	    {
	      rec->rec_buffer_data[index_rec+j]=com->buffer_data[index_acq+rec->channel_list[j]];
	    }
	  // move the rec_index and acq_index
	  index_rec+=rec->number_of_channels_to_save;
	  index_acq+=com->number_channels;
	}
      // can now free the comedi buffer for other thread
      pthread_mutex_unlock( &mutex_comedi_interface_buffer );
      // update variable to know how much new data the rec buffer contains
      rec->new_samples_in_buffer+=num_samples_to_transfer;
    }
  return 0;
}

int recording_interface_save_data(struct recording_interface* rec)
{
  int num_elements;
  num_elements=rec->new_samples_in_buffer*rec->number_of_channels_to_save;
  
  if (num_elements<=0)
    {
      return 0;
    }
  if(fwrite(rec->rec_buffer_data,sizeof(short int),num_elements, rec->file)!=num_elements)
    {
      fprintf(stderr,"problem saving data in recording_interface_save_data\n");
      fprintf(stderr,"recording should stop here\n");
      rec->is_recording=0;
      return -1;
    }
  rec->number_samples_saved+=rec->new_samples_in_buffer;
  rec->new_samples_in_buffer=0;
  return 0;
}

int recording_interface_close_file(struct recording_interface* rec)
{
  fprintf(stderr, "recording_interface_close_file()\n");
  fclose(rec->file);
  
  return 0;
}

int recording_interface_check_file_size(struct recording_interface* rec)
{
  fprintf(stderr, "recording_interface_check_file_size()\n");
  // get the size of the recorded file
  fseek(rec->file,0,SEEK_END);
  rec->file_size=ftell(rec->file);
  //  fseek(rec->file,0,ios::beg);
  
  // check if the size has the predicted size
  rec->predicted_file_size=(sizeof(short int)*rec->number_of_channels_to_save*rec->number_samples_saved);

  if(rec->predicted_file_size!=rec->file_size)
    {
      fprintf(stderr,"recording_interface_check_file_size(), the size of the recorded file is not what it should be, problem opening file\n");
      fprintf(stderr,"size predicted: %ld, observed size: %ld\n",rec->file_size,rec->predicted_file_size);

      return -1;
    }
  else
    {
      fprintf(stderr,"size predicted: %ld, observed size: %ld\n",rec->file_size,rec->predicted_file_size);
    }
  return 0;
}


int recording_interface_clear_current_recording_variables(struct recording_interface* rec)
{
  /* this reset some variables to 0. called at the end of a recording 
*/
  rec->number_samples_saved=0;
  rec->new_samples_in_buffer=0;
  return 0;
}

int recording_interface_get_channels_from_channel_store(struct recording_interface* rec,struct comedi_interface* com, GtkListStore  *store)
{
  // function to update the channel list of the recording interface from
  // the GtkListStore that is modified by user in the preferences dialog
  GtkTreeIter iter;
  gboolean enabled;
  int channel_index;
  int i,j;
  gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &iter);
  rec->number_of_channels_to_save=0;
  for (i=0; i < com->number_devices;i++)
    {
      for (j=0; j < com->dev[i].number_channels_analog_input; j++)
	{
	  gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, COL_NO, &channel_index, COL_SELECT, &enabled, -1);
	  if (enabled)
	    {
	      rec->channel_list[rec->number_of_channels_to_save]=channel_index;
	      rec->number_of_channels_to_save++;
	    }
	  gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &iter);
	}
    }
  return 0;
}
